/*-----------------------------------------------------------------------------
annot_ops.c

Shyam Prabhakar Sep 29 2003
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include<float.h>

#include "align_strucs.h"
#include "read_file.h"
#include "coord_map.h"
#include "classify_annots.h"

#define BUF_SIZ 1000
#define FEAT_ALLOC_SIZ 10


GFF_ANNOT transfer_annot(GFF_ANNOT source_annot, COORD_MAP *map_ptr,
			 char *target_seq_name)
{

  int i, nfeat, source_nfeat;
  int algn_start, algn_end, target_start, target_end;
  GFF_FEAT *feats, *source_feats;
  GFF_ANNOT target_annot;

  source_nfeat = source_annot.num_feat;
  source_feats = source_annot.feats;

  feats = (GFF_FEAT *) malloc(source_nfeat * sizeof(GFF_FEAT));

  nfeat = 0;
  for (i = 0; i < source_nfeat; i++)
    {

      algn_start = source_feats[i].algn_start;
      algn_end   = source_feats[i].algn_end;

      target_start = seq_coord(algn_start, '>', map_ptr);
      target_end   = seq_coord(algn_end,   '<', map_ptr);

      if (target_end >= target_start)
	{

	  strcpy(feats[nfeat].seqname, target_seq_name);
	  strcpy(feats[nfeat].source, source_feats[i].source);
	  strcpy(feats[nfeat].feature, source_feats[i].feature);

	  feats[nfeat].start = target_start;
	  feats[nfeat].end   = target_end;

	  feats[nfeat].algn_start = algn_start;
	  feats[nfeat].algn_end   = algn_end;

	  feats[nfeat].score = source_feats[i].score;

	  feats[nfeat].strand = source_feats[i].strand;
	  feats[nfeat].frame  = source_feats[i].frame;

	  strcpy(feats[nfeat++].attrib, source_feats[i].attrib);

	}
    }

  target_annot.feats    = feats;
  target_annot.num_feat = nfeat;
  return target_annot;

}


SEG_LIST annot_align_segs(GFF_ANNOT exon_annot, int len, char *region)
{

  int nseg, e, n_exon, seg_start, seg_end;
  GFF_FEAT *feats;
  SEG_LIST segs;

  n_exon = exon_annot.num_feat;
  feats  = exon_annot.feats;

  if ( !strcmp(region, "exon") )
    {
      segs.nseg   = n_exon;
      segs.starts = (int *) malloc(n_exon*sizeof(int));
      segs.ends   = (int *) malloc(n_exon*sizeof(int));

      for (e = 0; e < n_exon; e++)
	{
	  segs.starts[e] = feats[e].algn_start;
	  segs.ends[e]   = feats[e].algn_end;
	}
    }
  else if ( !strcmp(region, "nonexon") )
    {

      segs.starts = (int *) malloc( (n_exon + 1) *sizeof(int));
      segs.ends   = (int *) malloc( (n_exon + 1) *sizeof(int));

      nseg = seg_start = 0;      
      for (e = 0; e < n_exon; e++)
	{
	  if ( (seg_end = feats[e].algn_start - 1) >= seg_start )
	    {
	      segs.starts[nseg] = seg_start;
	      segs.ends[nseg++] = seg_end;
	    }

	  seg_start = feats[e].algn_end + 1;
	}

      if ( len > seg_start )
	{
	  segs.starts[nseg] = seg_start;
	  segs.ends[nseg++] = len - 1;
	}

      segs.nseg = nseg;

    }
  else
    {
      segs.nseg   = 1;
      segs.starts = (int *) malloc(sizeof(int));
      segs.ends   = (int *) malloc(sizeof(int));

      segs.starts[0] = 0;
      segs.ends[0]   = len - 1;
    }

  return segs;

}


void new_super_exon(GFF_FEAT *feat_ptr, int e_start, int e_end)
{

  strcpy(feat_ptr->seqname, "all_species");
  strcpy(feat_ptr->source, "exon_union");
  strcpy(feat_ptr->feature, "exon");

  feat_ptr->start = 0;
  feat_ptr->end   = 0;

  feat_ptr->algn_start = e_start;
  feat_ptr->algn_end   = e_end;

  feat_ptr->score  = 0.0;
  feat_ptr->strand = '+';
  feat_ptr->frame  = '.';

  strcpy(feat_ptr->attrib, "?");

}


GFF_ANNOT super_exons(GFF_ANNOT *annots, int nrow, int ncol)
{

  int i, j, k, nfeat, super_nfeat, len, in_exon;
  int f_start, f_end_plus1, sup_exon_start;
  char *annot_vec, *feat_name;
  GFF_ANNOT super_annot = {0};
  GFF_FEAT *feats, *super_feats;

  annot_vec    = (char *) calloc(ncol + 1, sizeof(int));
  super_feats = (GFF_FEAT *) malloc(0);

  for (i = 0; i < nrow; i++)
    {
      nfeat = annots[i].num_feat;
      feats = annots[i].feats;
      for (j = 0; j < nfeat; j++)
	{
	  feat_name = feats[j].feature;
	  if ( is_exon_cds_utr(feat_name) )
	    {
	      f_start     = feats[j].algn_start;
	      f_end_plus1 = feats[j].algn_end + 1;

	      for (k = f_start; k < f_end_plus1; annot_vec[k++] = '\1');
	    }
	}
    }

  super_nfeat = in_exon = 0;
  i = -1;

  while ( i++ < ncol )
    if ( in_exon )
      {
	if ( !annot_vec[i] )
	  {
	    if ( !( super_nfeat % FEAT_ALLOC_SIZ ) )
	      {
		len = super_nfeat + FEAT_ALLOC_SIZ;
		super_feats =
		  (GFF_FEAT *) realloc(super_feats, len * sizeof(GFF_FEAT));
	      }

	    new_super_exon(super_feats + super_nfeat++, sup_exon_start, i - 1);
	    in_exon = 0;
	  }
      }
    else if ( annot_vec[i] )
      {
	sup_exon_start = i;
	in_exon = 1;
      }

  free(annot_vec);

  super_annot.num_feat = super_nfeat;
  super_annot.feats    = super_feats;

  return super_annot;

}


void print_annots(GFF_ANNOT annot)
{

  GFF_FEAT feat, *feats = annot.feats;
  int i, num_feat = annot.num_feat;

  printf("\n");
  for (i = 0; i < num_feat; i++)
    {
      feat = feats[i];
      
/*       Start/end coords are incremented before printing because all */
/* 	 coordinates are zero-based within this program */
      printf("%s\t%s\t%s\t%d\t%d\t%.2e\t%c\t%c\t%s\n", feat.seqname,
	     feat.source, feat.feature, feat.start + 1, feat.end + 1,
	     feat.score, feat.strand, feat.frame, feat.attrib);
/*       printf("%s\t%s\t%s\t%d\t%d\t%d\t%d\t%c\t%s\n", feat.seqname, */
/* 	     feat.source, feat.feature, feat.start + 1, feat.end + 1, */
/* 	     feat.algn_start, feat.algn_end, feat.strand, feat.attrib); */
    }
  printf("\n");

}


/*-----------------------------------------------------------------------------
The comparison function used by qsort for sorting features in increasing order
of feature start coordinate.
-----------------------------------------------------------------------------*/
static int feat_compar(const void *a1, const void *a2)
{

  return ((GFF_FEAT *)a1)->start - ((GFF_FEAT *)a2)->start;

}


int read_GFF_feat(char *buffer, GFF_FEAT *feat_ptr, COORD_MAP *map_ptr,
		  int seq_len, int warning)
{

  static char **field_ptrs;
  static char scor_str[GFF_STR_SIZ+1], GFF_str_fmt[30];
  static int visit_counter;
  int feat_start, feat_end;

  if ( !visit_counter )
    {
      field_ptrs = (char **) calloc(9, sizeof(char *));
      sprintf(GFF_str_fmt, "%%%ds", GFF_STR_SIZ);

      visit_counter = 1;
    }

  if ( !get_str_fields(buffer, field_ptrs, 9) )  return 0;

  sscanf(field_ptrs[0], GFF_str_fmt, feat_ptr->seqname);
  sscanf(field_ptrs[1], GFF_str_fmt, feat_ptr->source);
  sscanf(field_ptrs[2], GFF_str_fmt, feat_ptr->feature);

  sscanf(field_ptrs[3], "%d", &feat_start);
  sscanf(field_ptrs[4], "%d", &feat_end);
  sscanf(field_ptrs[5], GFF_str_fmt, scor_str);

  if ( feat_end < 1 || feat_start > seq_len )
    {
      if ( !warning )
	fprintf(stderr, "WARNING: annotation outside sequence bounds\n");
      return 0;
    }

  if ( feat_start < 1 )      feat_start = 1;
  if ( feat_end > seq_len )  feat_end   = seq_len;

  /* NOTE: the  "-1" is because numbering in GFF files is usually
     one-based, and this program keeps all coordinates zero-based */
  feat_ptr->start = feat_start - 1;
  feat_ptr->end   = feat_end   - 1;

  feat_ptr->algn_start = align_coord(feat_ptr->start, map_ptr);
  feat_ptr->algn_end   = align_coord(feat_ptr->end,   map_ptr);

  if ( *scor_str == '.' )
    feat_ptr->score = FLT_MAX;
  else
    sscanf(scor_str, "%f", &feat_ptr->score);
 
  feat_ptr->strand = *field_ptrs[6];
  feat_ptr->frame  = *field_ptrs[7];

  strncpy(feat_ptr->attrib, field_ptrs[8], GFF_ATT_SIZ);
  chop_trail_spaces(feat_ptr->attrib);

  return 1;

}


void read_annots(char *fnames[], int nfile, ALIGNMENT *algnmnt_ptr)
{

  FILE *fp;
  char buffer[BUF_SIZ + 1], **seq_names, *feat_seq_nam;
  int warning, i, j, nrow, num_feat, seq_len;
  GFF_ANNOT *annot_array;
  GFF_FEAT *feats;
  COORD_MAP *map_ptr;

  nrow      = algnmnt_ptr->nrow;
  seq_names = algnmnt_ptr->seq_names;

  algnmnt_ptr->annots = annot_array =
    (GFF_ANNOT *) calloc(nrow, sizeof(GFF_ANNOT));

  for (i = 0; i < nfile; i++)
    {

      if ( !(fp = fopen(fnames[i], "r")) )
	{
	  fprintf(stderr, "couldn't open annotation file %s!\n", fnames[i]);
	  continue;
	}

      feat_seq_nam = 0;
      while ( fgets(buffer, BUF_SIZ + 1, fp) &&
	      !(feat_seq_nam = read_first_word(buffer, 'l')) ||
	      *buffer == '#' )  ;

      if ( !feat_seq_nam )
	{
	  fprintf(stderr, "empty annotation file: %s!\n", fnames[i]);
	  continue;
	}

      j = 0;
      while ( j < nrow && strcmp(feat_seq_nam, seq_names[j]) )  j++;

      if (j == nrow)
	{
	  fprintf(stderr, "file %s has annot. for unrecognised sequence: "
		  "%s!\n...ignoring this file\n", fnames[i], feat_seq_nam);
	  continue;
	}

      num_feat = 0;
      feats    = (GFF_FEAT *) malloc(FEAT_ALLOC_SIZ * sizeof(GFF_FEAT));

      map_ptr = algnmnt_ptr->coord_maps + j;
      seq_len = algnmnt_ptr->seq_lens[j];

      warning = 0;
      if ( !read_GFF_feat(buffer, feats + num_feat++, map_ptr, seq_len, 0) )
	{
	  fprintf(stderr, "WARNING: file %s: couldn't process line:\n%s"
		  "...further warnings on this file will be suppressed\n\n",
		  fnames[i], buffer);
	  warning = 1;

	  num_feat--;
	}

      while ( fgets(buffer, BUF_SIZ + 1, fp) )
	{

	  /* blank line or garbage characters? */
	  if ( !read_first_word(buffer, 'x') )  continue;

	  if ( !(num_feat % FEAT_ALLOC_SIZ) )
	    feats = (GFF_FEAT *)
	      realloc(feats, (num_feat + FEAT_ALLOC_SIZ) * sizeof(GFF_FEAT) );

	  if ( !read_GFF_feat(buffer, feats + num_feat++, map_ptr,
			      seq_len, warning) )
	    {
	      if ( !warning && (warning = 1) )
		fprintf(stderr, "WARNING: file %s: couldn't process line:\n%s"
			"...further warnings on this file will be suppressed"
			"\n\n",	fnames[i], buffer);

	      num_feat--;
	    }

	}

      /* sort features in increasing order of feature start coordinate */
      qsort(feats, num_feat, sizeof(GFF_FEAT), feat_compar);

      annot_array[j].feats    = feats;
      annot_array[j].num_feat = num_feat;

/*       print_annots(annot_array[j]); */
/*       exit(1); */

    }

  algnmnt_ptr->super_annot = super_exons(annot_array, nrow, algnmnt_ptr->ncol);

/*   print_annots(algnmnt_ptr->super_annot); */
/*   exit(1); */

}
