/*-----------------------------------------------------------------------------
gumby.c

Shyam Prabhakar Aug 04 2003
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <string.h>

#include "tree_ops.h"
#include "align_ops.h"
#include "parsim.h"
#include "annot_ops.h"
#include "mbgumbel.h"
#include "hit_list.h"
#include "prnt_gumby_matlab.h"
#include "arg_struc.h"

#define MAX_FNAME_LEN 100


void prnt_gumby(ALIGN_SCOR *scor_ptr, HIT1D *hits, int nhit,
		 ALIGNMENT *algnmnt_ptr, ARGS args, char *region)
{

  static FILE *fp_mat;
  FILE *fp_hits_align, *fp_hits_fasta;
  char fname[MAX_FNAME_LEN + 15];
  static int matlab_file_exists;
  char extn[30] = "_";
  char *out_fname = args.out_fname;

  if ( !out_fname )
    {
      prnt_aligned_hits(algnmnt_ptr, hits, nhit, region, stdout, args);
      return;
    }

  strcat(extn, region);

  fp_hits_align =
    fopen(strcat(strcat(strcpy(fname, out_fname), extn), ".align"), "w");
  prnt_aligned_hits(algnmnt_ptr, hits, nhit, region, fp_hits_align, args);
  fclose(fp_hits_align);

  fp_hits_fasta =
    fopen(strcat(strcat(strcpy(fname, out_fname), extn), ".mfa"), "w");
  prnt_fasta_hits(algnmnt_ptr, hits, nhit, fp_hits_fasta, args);
  fclose(fp_hits_fasta);

/*   if ( !strcmp(region, "all") ) */
/*     { */
/*       fp_mat = */
/* 	fopen(strcat(strcat(strcpy(fname, out_fname), extn), ".m"), "w"); */

/*       prnt_gumby_matlab(scor_ptr, hits, nhit, algnmnt_ptr, */
/* 			 args.project, fp_mat, args); */
/*       fclose(fp_mat); */
/*     } */
/*   else if (matlab_file_exists) */
/*     { */
/*       prnt_gumby_matlab(0, hits, nhit, algnmnt_ptr, args.project, fp_mat, */
/* 			 args); */
/*       fclose(fp_mat); */
/*     } */
/*   else */
/*     { */
/*       fp_mat = fopen(strcat(strcpy(fname, out_fname), ".m"), "w"); */

/*       prnt_gumby_matlab(scor_ptr, hits, nhit, algnmnt_ptr, */
/* 			 args.project, fp_mat, args); */

/*       matlab_file_exists = 1; */
/*     } */

}


void err_exit(char *msg, int exit_code)
{

  fprintf(stderr, "%s ... exiting!\n", msg);
  exit(exit_code);

}


char *cat_strings(int num_str, char *str_list[])
{

  int i, len = 0;
  char *out_str;

  if (num_str <= 0)  return 0;

  for (i = 0; i < num_str; i++)
    len += strlen(str_list[i]);

  out_str = (char *) calloc(len + num_str, sizeof(char));

  sprintf(out_str, "%s", str_list[0]);
  for (i = 1; i < num_str; i++)
    strcat(strcat(out_str, " "), str_list[i]);

  return out_str;

}


int scan_until_option(int num_arg, char *arg_list[], int i)
{

  for (  ; i < num_arg; i++)
    if ( arg_list[i][0] == '-' && !isdigit( arg_list[i][1] ) &&
	 arg_list[i][1] != '.' )  break;

   return i;

}


void prnt_usage(char *prog_name)
{

  fprintf(stderr, "Usage:\n%s tree_file alignment_file [options]\n\n",
	  prog_name);

  fprintf(stderr, "Options:\n"

	  "-annots\t gff_file1 gff_file2 ...\n"

	  "-ratio\t fast vs slow mismatch rate ratio (default: 1.5)\n"

	  "-minseq\t num_seq\n"
	  "\t alignment columns with at least num_seq A/G/C/T characters"
	  " are\n\t used to calculate the neutral (baseline) score"
	  " distribution\n\t (default: num_seq = largest value such that the"
	  " ratio of baseline\n\t \"neutral\" alignment columns to"
	  " base-sequence positions is at least 40%%)\n"

	  "-base\t seq_no\n\t base/reference sequence for pvalue calculations."
	  " High-scoring conserved\n\t segments are detected independent of"
	  " any particular reference sequence,\n\t but they are assigned"
	  " pvalues relative to a fixed length (plen) of base\n\t sequence."
	  " 0 <= seq_no < nseq, where nseq = # of sequences in the tree."
	  "\n\t (default: the leftmost sequence in the tree is the base)\n"

	  "-plen\t prob_len\n\t length of base sequence for which statistical"
	  " significance is\n\t calculated (default: 10 kb)\n"

	  "-prob\t probability_threshold (default: 0.5)\n"

	  "-blklen\t block_len\n\t min length of ungapped alignment block for"
	  " mismatch rate calculation\n\t (default 20)\n"

	  "-econs\t this flag directs the program to look for conserved "
	  "regions in exons\n\t & non-exonic regions separately\n"
	  "\t (default: look for conservation in the sequence as a whole)\n"

/* 	  "-proj\t seq_no\n\t sequence on whose coordinates the MATLAB" */
/* 	  " output is projected\n\t" */
/* 	  " (default: MATLAB output in alignment coords. (no projection))\n" */

	  "-o\t out_filename\n\n");

}


int parse_args(int num_arg, char *arg_list[], ARGS *args_ptr)
{

  int tmp_plen, j, i = 3;
  char *opt_str;
  double tmp_p;

  if (num_arg < 3)
    {
      prnt_usage(arg_list[0]);
      return 0;
    }

  args_ptr->tree_fname  = arg_list[1];
  args_ptr->algn_fname  = arg_list[2];
  args_ptr->mut_ratio   = 1.5;
  args_ptr->base        = 0;
  args_ptr->col_min_seq = 0;
  args_ptr->prob_thresh = 0.5;
  args_ptr->plen        = 1e4;
  args_ptr->blk_len     = 20;
  args_ptr->econs       = 0;
  args_ptr->project     = -1;
  args_ptr->invocation  = cat_strings(num_arg, arg_list);

  while ( (i = scan_until_option(num_arg, arg_list, i)) < num_arg )
    {

      opt_str = arg_list[i++] + 1;

      /* all flags require a value afterwards, except the "-econs" flag */
      if (i >= num_arg)
	if ( !strcmp(opt_str, "econs") )
	  {
	    args_ptr->econs = 1;
	    return 1;
	  }
	else
	  {
	    prnt_usage(arg_list[0]);
	    return 0;
	  }

      if ( !strcmp(opt_str, "ratio") )
	args_ptr->mut_ratio   = atof(arg_list[i++]);
      else if ( !strcmp(opt_str, "minseq") )
	args_ptr->col_min_seq = atoi(arg_list[i++]);
      else if ( !strcmp(opt_str, "prob") )
	{
	  tmp_p = atof(arg_list[i++]);

	  if (tmp_p > 0 && tmp_p <= 1)
	    args_ptr->prob_thresh = tmp_p;
	  else
	    fprintf(stderr, "\nInvalid probability threshold ... "
		    "using default value!\n\n");
	}
      else if ( !strcmp(opt_str, "plen") )
	{
	  if ( atof(arg_list[i]) > 0 )  args_ptr->plen = atof(arg_list[i++]);
	}
      else if ( !strcmp(opt_str, "blklen") )
	args_ptr->blk_len = atof(arg_list[i++]);
      else if ( !strcmp(opt_str, "o") )
	args_ptr->out_fname   = arg_list[i++];
      else if ( !strcmp(opt_str, "base") )
	args_ptr->base = atof(arg_list[i++]);
      else if ( !strcmp(opt_str, "proj") )
	args_ptr->project = atof(arg_list[i++]);
      else if ( !strcmp(opt_str, "econs") )
	args_ptr->econs = 1;
      else if ( !strcmp(opt_str, "annots") )
	{

	  j = scan_until_option(num_arg, arg_list, i);

	  if ( !(args_ptr->num_annot_fil = j-i) )
	    {
	      prnt_usage(arg_list[0]);
	      return 0;
	    }
	  else
	    {
	      args_ptr->annot_fnames = arg_list + i;
	      i = j;
	    }

	}
      else
	{
	  fprintf(stderr, "\noption \"%s\" not recognised!\n\n", opt_str);
	  prnt_usage(arg_list[0]);
	  return 0;
	}


    }

  return 1;

}


int main(int argc, char *argv[])
{

  PHYLO_TREE *phy_tree_ptr;
  ALIGNMENT *inp_algn_ptr, *tree_algn_ptr;
  ALIGN_SCOR col_scor = {0}, stat_scor = {0};
  GUMB_STATS gs;
  HIT1D *hit_list;
  int nhit;
  ARGS args = {0};

  if ( !parse_args(argc, argv, &args) )  err_exit("", 1);

  if ( !( phy_tree_ptr = read_tree(args.tree_fname) ) )
    err_exit("Couldn't read phylogenetic tree", 2);

  if ( !( inp_algn_ptr = read_align(args.algn_fname, "fasta") ) )
    err_exit("Couldn't read alignment", 3);

  if ( !( tree_algn_ptr = get_tree_align(phy_tree_ptr, inp_algn_ptr) ) )
    err_exit("Input alignment does not match tree", 4);

  read_annots(args.annot_fnames, args.num_annot_fil, tree_algn_ptr);

  if ( !parsim_log_like(tree_algn_ptr, args, &col_scor, &stat_scor) )
    err_exit("Couldn't calculate parsimony score", 2);  /* this won't happen */

  if ( !mbgumbel(&stat_scor, tree_algn_ptr, &gs, &args) )
    err_exit("", 6);

  if ( args.econs )
    {

      if ( hit_list =  find_1Dhits(&col_scor, tree_algn_ptr,
				   gs, args, &nhit, "nonexon") )
	{
	  prnt_gumby(&col_scor, hit_list, nhit, tree_algn_ptr, args,
		      "nonexon");
	  free_hit_list(hit_list, nhit);
	}

      if ( hit_list = find_1Dhits(&col_scor, tree_algn_ptr,
				  gs, args, &nhit, "exon") )
	{
	  prnt_gumby(&col_scor, hit_list, nhit, tree_algn_ptr, args, "exon");
	  free_hit_list(hit_list, nhit);
	}

    }
  else
    if ( hit_list = find_1Dhits(&col_scor, tree_algn_ptr,
				gs, args, &nhit, "all") )
      prnt_gumby(&col_scor, hit_list, nhit, tree_algn_ptr, args, "all");

  return 0;

}
