/*-----------------------------------------------------------------------------
hit_list.c

Shyam Prabhakar Oct 17 2003
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "scor_strucs.h"
#include "hit_strucs.h"
#include "mbgumbel.h"
#include "align_ops.h"
#include "coord_map.h"
#include "arg_struc.h"
#include "annot_ops.h"
#include "arg_struc.h"

#define HIT_ALLOC 10
#define PRNT_HIT_COLS 50
#define PRNT_FASTA_COLS 50


void hit_seq_coords(HIT1D *hit_ptr, int hstart, int hend, COORD_MAP *maps,
		    int nrow)
{

  int i;

  hit_ptr->seq_starts = (int *) malloc(nrow * sizeof(int));
  hit_ptr->seq_ends   = (int *) malloc(nrow * sizeof(int));
  
  for (i = 0; i < nrow; i++)
    {
      hit_ptr->seq_starts[i] = seq_coord(hstart, '>', maps + i);
      hit_ptr->seq_ends[i]   = seq_coord(hend,   '<', maps + i);
    }

}


HIT1D *find_1Dhits(ALIGN_SCOR *scor_ptr, ALIGNMENT *algnmnt_ptr, GUMB_STATS gs,
		   ARGS args, int *nhit, char *region)
{

  int hit_len, scor_len, j;
  int hit_scor, hit_start, hit_end, cum_sum, nh, scor_lim;
  int nrow, *scor_vec, plen;
  int seg, nseg, seg_end_plus1, *seg_starts, *seg_ends;
  HIT1D *hits;
  double pthresh, lambda, K;
  COORD_MAP *coord_maps;
  SEG_LIST target_segs;

  if ( !(nrow = algnmnt_ptr->nrow) )  return 0;
  coord_maps = algnmnt_ptr->coord_maps;

  scor_vec = scor_ptr->scor;
  scor_len = scor_ptr->len;

  lambda = gs.lambda;
  K      = gs.K;

  /* calculate score threshold correponding to probability threshold */
  /* plen is the length of alignment for which a statistical significance is
     calculated. */

  pthresh = args.prob_thresh;
  plen    = (int) (gs.base_ratio * args.plen);

  scor_lim = floor( pval2scor(gs, pthresh, plen) );

  if (scor_lim < 0)  scor_lim = 0;

  hits = (HIT1D *) malloc(0);
  nh   = 0;

  target_segs = annot_align_segs(algnmnt_ptr->super_annot, scor_len, region);
  nseg        = target_segs.nseg;
  seg_starts  = target_segs.starts;
  seg_ends    = target_segs.ends;

  for (seg = 0; seg < nseg; seg++)
    {

      hit_start     = seg_starts[seg] - 1;
      seg_end_plus1 = seg_ends[seg] + 1;

      while ( ++hit_start < seg_end_plus1 )
	{

	  if ( (cum_sum = scor_vec[hit_start]) <= 0)  continue;

	  hit_scor = cum_sum;
	  hit_end  = hit_start;

	  for (j = hit_start + 1; cum_sum > 0 && j < seg_end_plus1; j++)
	    if ( (cum_sum += scor_vec[j]) > hit_scor )
	      {
		hit_scor = cum_sum;  /* score of segment */
		hit_end  = j;
	      }

	  if (hit_scor > scor_lim)
	    {

	      if ( !(nh % HIT_ALLOC) )
		hits = (HIT1D *) realloc(hits, (nh + HIT_ALLOC)*sizeof(HIT1D));

	      hits[nh].start = hit_start;
	      hits[nh].len   = hit_end - hit_start + 1;

	      hit_seq_coords(hits + nh, hit_start, hit_end, coord_maps, nrow);

	      hits[nh].scor   = hit_scor;
	      hits[nh++].pval = pvalue(lambda, K, plen, hit_scor);

	      hit_start = hit_end + 1;

	    }
	  else
	    hit_start = j - 1;

	}

    }

  if (*nhit = nh)
    return hits;
  else
    return 0;

}


void prnt_hit_align(char **align_matrx, int nrow, char **seq_names,
		    int name_len, HIT1D hit, FILE *fp)
{

  int row, i;
  int h_start, h_end, h_end_prev, tmp_len, coord_len, *seq_starts, *seq_ends;
  static char tmp_str[100];

  h_start    = hit.start;
  h_end      = h_start + hit.len;
  h_end_prev = h_end - PRNT_HIT_COLS;
  seq_starts = hit.seq_starts;
  seq_ends   = hit.seq_ends;

  /* one is added to seq_end coord. to make it one-based in output */
  for (coord_len = 0, row = 0; row < nrow; row++)
    if ( (tmp_len = sprintf(tmp_str, "%d", seq_ends[row] + 1)) > coord_len)
      coord_len = tmp_len;

  for (i = h_start; i < h_end_prev; i += PRNT_HIT_COLS)
    {

      for (row = 0; row < nrow; row++)
	{
	  fprintf(fp, "%-*s %*d %*d  ", name_len, seq_names[row], coord_len,
		  seq_starts[row] + 1, coord_len, seq_ends[row] + 1);
	  fprintf(fp, "%.*s\n", PRNT_HIT_COLS, align_matrx[row] + i);
	}
      fprintf(fp, "\n");

    }

  for (row = 0; row < nrow; row++)
    {
      fprintf(fp, "%-*s %*d %*d  ", name_len, seq_names[row], coord_len,
	      seq_starts[row] + 1, coord_len, seq_ends[row] + 1);
      fprintf(fp, "%.*s\n", h_end - i, align_matrx[row] + i);
    }
  fprintf(fp, "\n");

}


int max_strlen(char **str_array, int nstr)
{

  int i, len, max_len = 0;

  for (i = 0; i < nstr; i++)
    if ( (len = strlen(str_array[i])) > max_len)
      max_len = len;

  return max_len;

}


void prnt_aligned_hits(ALIGNMENT *algnmnt_ptr, HIT1D *hits, int nhit,
		       char *region, FILE *fp, ARGS args)
{

  int i, name_len, nrow, ncol, h_start, hit_len;
  char **seq_names, **matrix;
  HIT1D hit;

  ncol      = algnmnt_ptr->ncol;
  nrow      = algnmnt_ptr->nrow;
  matrix    = algnmnt_ptr->matrix;
  seq_names = algnmnt_ptr->seq_names;

  name_len  = max_strlen(seq_names, nrow);

  fprintf(fp, "%s\n\n", args.invocation);
  fprintf(fp, "%s: no. of hits = %d\n", region, nhit);

  for (i = 0; i < nhit; i++)
    {

      hit     = hits[i];
      h_start = hit.start;
      hit_len = hit.len;

      if ( (h_start + hit_len) > ncol )
	{
	  fprintf(stderr, "hit no. %d extends beyond end of alignment ... "
		  "cannot print!\n", i);
	  return;
	}

      fprintf(fp, "\nstart (algnmnt coords, 0-based) %d, len %d, score %d, "
	      "pvalue %.2g\n", h_start, hit_len, hit.scor, hit.pval);
      prnt_hit_align(matrix, nrow, seq_names, name_len, hit, fp);

    }

}


void prnt_hit_fasta(char **align_matrx, int nrow, char **seq_names, HIT1D hit,
		    int max_hit_len, FILE *fp)
{

  int row, i;
  int h_start, h_len, seq_len, *seq_starts, *seq_ends;
  char *algn_start, *algn_end, *blk_start, *blk_end;
  static int old_mh_len, is_gap[256] = {0};
  static char *seq;

  if ( !seq )
    {
      identify_gap_chars(is_gap);
      seq = (char *) malloc( (max_hit_len + 1) * sizeof(char));
      old_mh_len = max_hit_len;
    }
  else if ( old_mh_len < max_hit_len )
    {
      seq = (char *) realloc(seq, (max_hit_len + 1) * sizeof(char));
      old_mh_len = max_hit_len;
    }

  h_len      = hit.len;
  h_start    = hit.start;
  seq_starts = hit.seq_starts;
  seq_ends   = hit.seq_ends;

  for (row = 0; row < nrow; row++)
    {

      /* one is added to make seq. coords. one-based in the output */
      fprintf(fp, ">%s %d %d, score %d pvalue %.2g\n", seq_names[row],
	      seq_starts[row] + 1, seq_ends[row] + 1, hit.scor, hit.pval);

      algn_start = align_matrx[row] + h_start;
      algn_end   = algn_start + h_len;

      *seq = '\0';

      if ( is_gap[*algn_start] )
	{
	  if ( blk_start = strpbrk(algn_start, ">*") )  blk_start++;
	}
      else
	blk_start = algn_start;

      while (blk_start && blk_start < algn_end)
	{
	  
	  if ( !(blk_end = strpbrk(blk_start, "<*")) || blk_end > algn_end )
	    {
	      strncat(seq, blk_start, algn_end - blk_start);
	      break;
	    }

	  strncat(seq, blk_start, blk_end - blk_start);

	  if ( blk_start = strpbrk(blk_end, ">*") )  blk_start++;

	}

      seq_len = strlen(seq);
      for (i = 0; i < seq_len; i += PRNT_FASTA_COLS)
	fprintf(fp, "%.*s\n", PRNT_FASTA_COLS, seq + i);
      fprintf(fp, "\n");

    }

}


void prnt_fasta_hits(ALIGNMENT *algnmnt_ptr, HIT1D *hits, int nhit, FILE *fp,
		     ARGS args)
{

  int i, nrow, ncol, hit_len, max_hit_len;
  char **seq_names, **matrix;
  HIT1D hit;

  ncol      = algnmnt_ptr->ncol;
  nrow      = algnmnt_ptr->nrow;
  matrix    = algnmnt_ptr->matrix;
  seq_names = algnmnt_ptr->seq_names;

  /* length of longest hit in alignment coordinates */
  for (max_hit_len = 0, i = 0; i < nhit; i++)
    if ( (hit_len = hits[i].len) > max_hit_len )
      max_hit_len = hit_len;

  fprintf(fp, "%s\n\n", args.invocation);
  for (i = 0; i < nhit; i++)
    {

      hit = hits[i];

      if ( (hit.start + hit.len) > ncol )
	{
	  fprintf(stderr, "hit no. %d extends beyond end of alignment ... "
		  "cannot print!\n", i);
	  return;
	}

      prnt_hit_fasta(matrix, nrow, seq_names, hit, max_hit_len, fp);

    }

}


void free_hit_list(HIT1D *list, int nhit)
{

  int i;

  if ( !nhit )  return;

  for (i = 0; i < nhit; i++)
    {
      free(list[i].seq_starts);
      free(list[i].seq_ends);
    }

  free(list);

}
